
/* KI4LKF */

#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#define VERSION "1.15"
#define IP_SIZE 15

static unsigned char recv_buf[2000];

static sockaddr_in g2_internal;
static sockaddr_in mod_ctrl;

/* the multiple repeater bands */
static const short int MAX_MODS=3;
static short int MODS=0;
static struct
{
   struct sockaddr_in info;
   char type;  /* R=RF user can talk, T=RF user can copy, B=Both */
} band[MAX_MODS]; 

static fd_set fdset;
static struct timeval tv;
static int insock = -1;

static bool keep_running = true;

static const int TRACE_BFSZ = 512;
static char trace_buf[TRACE_BFSZ];

static int read_config(const char *cfgFile);
static void traceit(const char *fmt,...);
static void sigCatch(int signum);

/* signal catching function */
static void sigCatch(int signum)
{
   /* do NOT do any serious work here */
   if ((signum == SIGTERM) ||
       (signum == SIGINT))
      keep_running = false;
   return;
}

static int read_config(const char *cfgFile)
{
   short int valid_params = 3;
   short int params = 0;

   FILE *cnf = NULL;
   char inbuf[1024];
   char *p = NULL;
   char *ptr;

   const char *delim = ":";
   char *ip = NULL;
   char *port = NULL;
   char *type = NULL;

   cnf = fopen(cfgFile, "r");
   if (!cnf)
   {
      traceit("Failed to open file %s\n", cfgFile);
      return 1;
   }

   traceit("Reading file %s\n", cfgFile);
   while (fgets(inbuf, 1020, cnf) != NULL)
   {
      if (strchr(inbuf, '#'))
         continue;

      p = strchr(inbuf, '\r');
      if (p)
         *p = '\0';
      p = strchr(inbuf, '\n');
      if (p)
         *p = '\0';

      p = strchr(inbuf, '=');
      if (!p)
         continue;
      *p = '\0';

      if (strcmp(inbuf,"BAND") == 0)
      {
         ip = NULL; port = NULL;
         traceit("BAND=[%s]\n", p + 1);
         if (MODS < MAX_MODS)
         {
            ptr = strchr(p + 1, ' ');
            if (ptr) 
               *ptr = '\0';
            
            if (strlen(p + 1) < 1)
               traceit("BAND value [%s] invalid\n", p + 1);
            else
            {
               ip = strtok(p + 1, delim);
               port = strtok(NULL, delim);
               type = strtok(NULL, delim);

               if ((ip == NULL) || (port == NULL) || (type == NULL))
                  traceit("BAND info invalid\n");
               else
               {
                  traceit("   IP=[%s],port=[%s], type=[%s]\n", ip, port, type);
                  memset(&(band[MODS]), 0, sizeof(band[MODS]));
                  band[MODS].info.sin_family = AF_INET;
                  band[MODS].info.sin_port = htons(atoi(port));
                  band[MODS].info.sin_addr.s_addr = inet_addr(ip);
                  band[MODS].type = *type;
                  MODS++;
                  if (MODS == 1)
                     params ++;
               }
            }
         }
         else
            traceit("Warning:MAX BANDs is %d, [%s] will not be processed\n", MAX_MODS, p + 1);
      }
      else
      if (strcmp(inbuf,"G2_INTERNAL") == 0)
      {
         ip = NULL; port = NULL;
         traceit("G2_INTERNAL=[%s]\n", p + 1);
         ptr = strchr(p + 1, ' ');
         if (ptr)
            *ptr = '\0';

         if (strlen(p + 1) < 1)
            traceit("G2_INTERNAL value [%s] invalid\n", p + 1);
         else
         {
            ip = strtok(p + 1, delim);
            if (ip)
               port = strtok(NULL, delim);

            if ((ip == NULL) || (port == NULL))
               traceit("G2_INTERNAL info invalid\n");
            else
            {
               traceit("   IP=[%s],port=[%s]\n", ip, port);
               memset(&g2_internal, 0, sizeof(g2_internal));
               g2_internal.sin_family = AF_INET;
               g2_internal.sin_port = htons(atoi(port));
               g2_internal.sin_addr.s_addr = inet_addr(ip);
               params ++;
            }
         }
      }
      else
      if (strcmp(inbuf,"MOD_CTRL") == 0)
      {
         ip = NULL; port = NULL;
         traceit("MOD_CTRL=[%s]\n", p + 1);
         ptr = strchr(p + 1, ' ');
         if (ptr)
            *ptr = '\0';

         if (strlen(p + 1) < 1)
            traceit("MOD_CTRL value [%s] invalid\n", p + 1);
         else
         {
            ip = strtok(p + 1, delim);
            if (ip)
               port = strtok(NULL, delim);

            if ((ip == NULL) || (port == NULL))
               traceit("MOD_CTRL info invalid\n");
            else
            {
               traceit("   IP=[%s],port=[%s]\n", ip, port);
               memset(&mod_ctrl, 0, sizeof(mod_ctrl));
               mod_ctrl.sin_family = AF_INET;
               mod_ctrl.sin_port = htons(atoi(port));
               mod_ctrl.sin_addr.s_addr = inet_addr(ip);
               params ++;
            }
         }
      }
   }
   fclose(cnf);

   if (params != valid_params)
   {
      traceit("Configuration file %s invalid\n",cfgFile);
      return 1;
   }
   return 0;
}

static void traceit(const char *fmt,...)
{
   time_t ltime;
   struct tm mytm;

   time(&ltime);
   localtime_r(&ltime,&mytm);

   snprintf(trace_buf,TRACE_BFSZ - 1,"%02d%02d%02d at %02d:%02d:%02d:",
            mytm.tm_mon+1,mytm.tm_mday,mytm.tm_year % 100,
            mytm.tm_hour,mytm.tm_min,mytm.tm_sec);

   va_list args;
   va_start(args,fmt);
   vsnprintf(trace_buf + strlen(trace_buf), TRACE_BFSZ - strlen(trace_buf) - 1, fmt, args);
   va_end(args);

   fprintf(stdout, "%s", trace_buf);
   return;
}

int main(int argc, char *argv[])
{
   struct sigaction act;
   short i,j;
   struct  sockaddr_in from;
   socklen_t fromlen;
   int len;
   int rc;
   char temp_mod = ' ';

   setvbuf(stdout, NULL, _IOLBF, 0);
   traceit("mod_ctrl VERSION %s\n", VERSION);

   if (argc != 2)
   {
      traceit("Usage: mod_ctrl mod_ctrl.cfg\n");
      return 1;
   }

   act.sa_handler = sigCatch;
   sigemptyset(&act.sa_mask);
   act.sa_flags = SA_RESTART;
   if (sigaction(SIGTERM, &act, 0) != 0)
   {
      traceit("sigaction-TERM failed, error=%d\n", errno);
      return 1;
   }
   if (sigaction(SIGINT, &act, 0) != 0)
   {
      traceit("sigaction-INT failed, error=%d\n", errno);
      return 1;
   }

   rc = read_config(argv[1]);
   if (rc != 0)
   {
      traceit("Failed to process config file %s\n", argv[1]);
      return 1;
   }

   insock = socket(PF_INET, SOCK_DGRAM, 0);
   if (insock == -1)
   {
      traceit("Failed to create insock, error=%d, message=%s\n",errno,strerror(errno));
      return 1;
   }
   rc = bind(insock, (struct sockaddr *)&mod_ctrl, sizeof(mod_ctrl));
   if (rc == -1)
   {
      traceit("Failed to bind server socket, error=%d, message=%s\n", errno,strerror(errno));
      close(insock); insock = -1;
      return 1;
   }
   fcntl(insock,F_SETFL,O_NONBLOCK);

   traceit("Entering processing loop...\n");
   while (keep_running)
   {
      tv.tv_sec = 0;
      tv.tv_usec = 20000;
      fromlen = sizeof(struct sockaddr);

      FD_ZERO(&fdset);
      FD_SET(insock,&fdset);
      tv.tv_sec = 0;
      tv.tv_usec = 20000; 
      (void)select(insock + 1,&fdset,0,0,&tv);

      if (FD_ISSET(insock, &fdset))
      {
         len = recvfrom (insock, (char *)recv_buf, 2000, 0, (struct sockaddr *)&from, &fromlen);

         /* lets see if we received from a local band */
         for (i = 0; i < MODS; i++)
         {
            if ((from.sin_addr.s_addr == band[i].info.sin_addr.s_addr) &&
                (from.sin_port == band[i].info.sin_port) &&
                ((band[i].type == 'R') || (band[i].type == 'B')))
               break;
         }
         if (i < MODS)
         { 
            /* we received from a repeater module */

            /* send the data to the gateway */
            sendto(insock, (char *)recv_buf, len, 0, (struct sockaddr *)&g2_internal, sizeof(g2_internal));

            /* send the data to the other local repeaters */
            if (len == 58)
            {
               temp_mod = recv_buf[35];
               recv_buf[35] = recv_buf[27];
               recv_buf[27] = temp_mod;       
            }
            else
            if (len == 32)
               len = 29;

            for (j = 0; j < MODS; j++)
            {
               if (i != j) 
               {
                  if ((band[j].type == 'T') || (band[j].type == 'B'))
                     sendto(insock, (char *)recv_buf, len, 0, (struct sockaddr *)&(band[j].info), sizeof(band[j].info));
               }
            }
         }
         else
         /* maybe we received data from the gateway */
         if ((from.sin_addr.s_addr == g2_internal.sin_addr.s_addr) &&
             (from.sin_port == g2_internal.sin_port))
         {
            /* 
               We received data from the internal IP of G2.
               Send the data to all the local bands
            */
            for (i = 0; i < MODS; i++)
            {
               if ((band[i].type == 'T') || (band[i].type == 'B'))
                  sendto(insock,(char *)recv_buf, len, 0, 
                         (struct sockaddr *)&(band[i].info), 
                         sizeof(band[i].info));
            }
         }
         FD_CLR (insock,&fdset);
      }
   }
   close(insock);

   return 0;
}

